/*
 * Individual.cpp
 *
 *  Created on: 2013-03-15
 *      Author: Kamil
 */



#include "Individual.h"
mt19937 generator(chrono::system_clock::now().time_since_epoch().count());

// Construct an Individual with the specified parameters
Individual::Individual(int size, double mrate, double prate)
:_age(0)
{
	uniform_real_distribution<double> dist(-32.768, 32.768);

	size = setSize(size);
	for(int i = 0; i < size; i++)
		_genotype.push_back(dist(generator));

	_elite = false;
	setMR(mrate);
	setPR(prate);
	setSDF(0.1);

	setFitness();
}

// Copy Constructor
Individual::Individual(const Individual &other)
{
	_elite = other._elite;
	_age = other._age;

	_genotype=other._genotype;

	_mrate = other._mrate;
	_prate = other._prate;
	_sdfactor = other._sdfactor;
	_fitness = other._fitness;
}

// Constructor with pointer parameter
Individual::Individual(const Individual* other)
{
	_elite = *(new double(other->_elite));
	_age = *(new double(other->_age));
	_genotype=other->_genotype;
	_mrate = *(new double(other->_mrate));
	_prate = *(new double(other->_prate));
	_sdfactor = *(new double(other->_sdfactor));
	_fitness = *(new double(other->_fitness));
}

// Empty destructor - no dynamically allocated memory
Individual::~Individual()
{

}

// Assignment
Individual& Individual::operator =(const Individual& other)
{
	_elite = other._elite;
	_age = other._age;

	_genotype=other._genotype;

	_mrate = other._mrate;
	_prate = other._prate;
	_sdfactor = other._sdfactor;
	_fitness = other._fitness;

	return *this;
}

// Assignment with pointer parameter
Individual* Individual::operator =(const Individual* other)
{
	_elite = other->_elite;
	_age = other->_age;
	_genotype=other->_genotype;
	_mrate = other->_mrate;
	_prate = other->_prate;
	_sdfactor = other->_sdfactor;
	_fitness = other->_fitness;

	return this;
}

// Set the size of the Individual, making sure the size is valid
int Individual::setSize(int size)
{
	uniform_int_distribution<int> dist(1, defaultMaxSize);

	if(size < 1)
		size = dist(generator);

	return size;
}

// set the elite status of the Individual
void Individual::setElite()
{
	_elite = true;
}

// Calculate fitness
void Individual::setFitness()
{
	double t1 = 0.0, t2 = 0.0;

	for(int i = 0; i < size(); i++)
	{
		t1 += pow(_genotype[i], 2);
		t2 += cos(2 * M_PI * _genotype[i]);
	}

	t1 = t1 / static_cast<double>(size());
	t1 = -0.2 * sqrt(t1);
	t1 = -20 * pow(M_E, t1);
	t2 = t2 / static_cast<double>(size());
	t2 = pow(M_E, t2);

	_fitness = t1 - t2 + 20 + M_E;
}

// Mutate the Individual
void operator!(Individual& I)
{
	uniform_real_distribution<double> uDist(0.0, 1.0);
	uniform_real_distribution<double> rDist(-32.768, 32.768);
	double sdf = I._sdfactor;
	double mrate = I._mrate;
	double mean = 0.0, sd = 0.0, newValue = 0.0, prob = 0.0;

	for(int i = 0; i < I.size(); i++)
	{
		prob = uDist(generator);
		if(prob < mrate && !I.isElite())
		{
			mean = I._genotype[i];
			sd = mean * sdf;
			normal_distribution<double> nDist(mean, sd);
			nDist.reset();
			newValue = nDist(generator);
			if(newValue > 32.768)
				newValue = 32.768;
			else if(newValue < -32.768)
				newValue = -32.768;
			I._genotype[i] = newValue;
		}
	}
	I.setFitness();
	I._age++;
}

// Set the Individual's mutation rate
void Individual::setMR(double mrate)
{
	if(mrate < 0.0)
		mrate = 0.0;
	else if(mrate > 1.0)
		mrate = 1.0;
	else
		_mrate = mrate;
}

// Set the Individual's permutation rate
void Individual::setPR(double prate)
{
	if(prate < 0.0)
		prate = 0.0;
	else if(prate > 1.0)
		prate = 1.0;
	else
		_prate = prate;
}

// Set the SD factor used in mutation
void Individual::setSDF(double sdf)
{
	if(sdf < 0.0)
		sdf = 0.0;
	else if(sdf > 1.0)
		sdf = 1.0;
	else
		_sdfactor = sdf;
}

// Return fitness
double Individual::fitness()
{
	return _fitness;
}

// Return mutation rate
double Individual::mutationRate()
{
	return _mrate;
}

// return permuation rate
double Individual::permutationRate()
{
	return _prate;
}

// return SD factor used in mutation
double Individual::sdFactor()
{
	return _sdfactor;
}

// return Individual size
int Individual::size()
{
	return _genotype.size();
}

// return Individual age
int Individual::age()
{
	return _age;
}

// return elite status
bool Individual::isElite()
{
	return _elite;
}

// return the allele at the given position
double Individual::allele(int position)
{
	if (position < 0 || position > size() - 1)
		throw "Out of Bounds";
	else
		return _genotype[position];
}

string Individual::info()
{
	stringstream output;
	output << "[";
	for(int i = 0; i < size(); i++)
	{
		output << " " << _genotype[i] << " ";
		if(i != size() - 1)
			output << "|";
	}
	output << "]";
	return output.str();
}

vector<Individual> operator*(Individual& p1, Individual& p2)
{
	uniform_real_distribution<double> uDist(0.0, 1.0);
	double a = 0.0;
	int smaller = 0;

	double prob = uDist(generator);

	if(abs(p1.permutationRate() - prob) < abs(p2.permutationRate() - prob))
		a = p1.permutationRate();
	else
		a = p2.permutationRate();

	double b = 1 - a;

	if(p1.size() < p2.size())
		smaller = p1.size();
	else
		smaller = p2.size();

	double mr = (p1.mutationRate() + p2.mutationRate()) / 2;
	double pr = (p1.permutationRate() + p2.permutationRate()) / 2;
	double sdf = (p1.sdFactor() + p2.sdFactor()) / 2;

	Individual c1(smaller);
	Individual c2(smaller);

	for(int i = 0; i < smaller; i++)
	{
		c1._genotype[i] = (a * p1.allele(i)) + (b * p2.allele(i));
		c2._genotype[i] = (b * p1.allele(i)) + (a * p2.allele(i));
	}

	c1.setSDF(sdf);
	c2.setSDF(sdf);
	c1.setMR(mr);
	c2.setMR(mr);
	c1.setPR(pr);
	c2.setPR(pr);

	vector<Individual> container;
	container.push_back(c1);
	container.push_back(c2);

	return container;
}

vector<Individual> operator+(Individual& parent1, Individual& parent2)
{
	double a = 0.0;
	uniform_real_distribution<double> dist(0.0, 1.0);
	double prob = dist(generator);

	if(abs(parent1.permutationRate() - prob) < abs(parent2.permutationRate() - prob))
	{
		a = parent1.permutationRate();
	}
	else
		a = parent2.permutationRate();

	int counter = 0;
	double temp = 0.0;
	Individual c1(parent1.size());
	Individual c2(parent2.size());
	while(counter < c1.size() && counter < c2.size())
	{
		prob = dist(generator);
		c1._genotype[counter] = parent1._genotype[counter];
		c2._genotype[counter] = parent2._genotype[counter];
		if(prob < a)
		{
			temp = c1._genotype[counter];
			c1._genotype[counter] = c2._genotype[counter];
			c2._genotype[counter] = temp;
		}
		counter++;
	}

	vector<Individual> container;
	container.push_back(c1);
	container.push_back(c2);

	return container;
}
