/*
 * Population.cpp
 *
 *  Created on: 2013-03-16
 *      Author: Kamil
 */




#include "Population.h"
mt19937 generator2(chrono::system_clock::now().time_since_epoch().count());

// Constructor pushes the desired number of Individuals into the pool
// Remember to update Population statistics
Population::Population(int size, int indsize)
{
	double tolerance = 0.00001;
	size = setSize(size);

	for(int i = 0; i < size; i++)
	{
		Individual* a = new Individual(indsize);
		pool.push_back(*a);
	}
	update();
}

// Copy constructor
Population::Population(const Population& that)
{
	pool=that.pool;
	_age = that._age;
	_fitness = that._fitness;
	_tfitness = that._tfitness;
	_stddev = that._stddev;
	_rsd = that._rsd;
	elites=that.elites;
	tolerance = that.tolerance;
}

// Destructor clears the pool
Population::~Population()
{
	pool.clear();
}

// Assignment
Population& Population::operator=(const Population& that)
{
	pool=that.pool;
	_age = that._age;
	_fitness = that._fitness;
	_tfitness = that._tfitness;
	_stddev = that._stddev;
	_rsd = that._rsd;
	elites = (that.elites);
	tolerance = that.tolerance;

	return *this;
}

// Sets the size of the Population, making sure that the desired size is valid
int Population::setSize(int size)
{
	uniform_int_distribution<int> dist(1, defaultMaxSize);
	if(size < 0)
		size = dist(generator2);
	return size;
}

// Updates Population statistics
void Population::update()
{
	double sumfitness = 0.0, sumvar = 0.0, var = 0.0;
	int sumage = 0;
	for(int i = 0; i < size(); i++)
	{
		sumfitness += pool[i].fitness();
		sumage += pool[i].age();
	}
	_tfitness = sumfitness;
	_fitness = _tfitness / (double) size();
	_age = (double) sumage / (double) size();

	for(int i = 0; i < size(); i++)
	{
		var = pool[i].fitness() - _fitness;
		var = var * var;
		sumvar = sumvar + var;
	}

	_stddev = sqrt(sumvar / size());
	_rsd = abs(_stddev / _fitness);
}

// Fitness-proportional rank of the Individual
double Population::fprank(Individual& target)
{
	return target.fitness() / _tfitness;
}

// Disqualify Individuals older than the desired age limit
void Population::disqualify(double limit)
{
	if(limit > 1.0)
	{
		vector<Individual>::iterator itr;
		for(itr = pool.begin(); itr != pool.end(); itr++)
		{
			if((double) (*itr).age() > limit && !(*itr).isElite())
				pool.erase(itr);
		}
	}
	update();
}

// Set the individual's elite status
void Population::setElite(Individual& newElite)
{
	elites.push_back(newElite);
	newElite.setElite();
}

// Roulette selection
// Resulting size is the same as current population size
void Population::roulette()
{
	// Create an array for the new pool
	vector<Individual> newpool;

	uniform_real_distribution<double> dist(0.0, _tfitness);
	double prob = 0.0;

	// Immediately push all elites into the new pool
	for(int i = 0; i < numElites(); i++)
		newpool.push_back(elites[i]);

	// To fill the remaining spaces
	int counter = 0;
	while(newpool.size() < pool.size())
	{
		// Spin the wheel
		prob = dist(generator2);
		counter = 0;
		// Find the Individual that it is pointing to
		while(prob > tolerance)
		{
			prob = prob - pool[counter].fitness();
			counter++;
		}
		// And push the Individual into the new pool
		newpool.push_back(pool[counter - 1]);
	}

	// Set the population pool = the new pool
	pool = newpool;
	// Update the statistics
	update();
}

// Tournament selection
// Input is the sample size
// Window size is fitness-proportional
void Population::tournament(int samplesize)
{
	// Figure out the proportion and multiply size by proportion
	// to get the window size
	double proportion = fittest().fitness() / fitness();
	int windowsize = (int) (size() * proportion);

	// Make sure the window size is valid; it needs to be at least the
	// greater of 2 and number of elite Individuals
	if(windowsize < 2)
		windowsize = 2;
	else if(windowsize < numElites())
		windowsize = numElites();
	else if(windowsize > size())
		windowsize = size();

	vector<Individual> newpool;

	if(samplesize == -1)
		samplesize = windowsize;
	else if(samplesize < 0)
		samplesize = 0;
	else if(samplesize > size())
		samplesize = size();

	vector<Individual> minipool;
	uniform_int_distribution<int> sampling(0, size() - 1);

	// Save all the elites into the new pool
	vector<Individual>::iterator itr1;
	for(itr1 = elites.begin(); itr1 != elites.end(); itr1++)
		newpool.push_back(*itr1);

	// While spaces remain in the new pool
	int index = 0;
	while(newpool.size() < windowsize)
	{
		// Pick up a handful of Individuals (samplesize)
		for(int i = 0; i < samplesize; i++)
		{
			index = sampling(generator2);
			minipool.push_back(pool[index]);
		}
		// And save the best of them into the new pool
		newpool.push_back(fittest(minipool));
		minipool.clear();
	}

	// Set the pool = new pool and update statistics
	pool = newpool;
	update();
}

// Mutate the entire Population
void Population::mutate()
{
	mutate(0, size() - 1);
}

// Mutate Individuals from index min to index max (inclusive)
void Population::mutate(int min, int max)
{
	int temp = 0;
	if(min < 0)
		min = 0;
	else if(max > size() - 1)
		max = size() - 1;

	if(max < min)
	{
		temp = max;
		max = min;
		min = temp;
	}

	// call the mutation function for the appropriate Individuals
	for(int i = min; i <= max; i++)
		!pool[i];

	// Update statistics
	update();
}

// Crossover for all individuals
// The switch indicates whether to use linear combination or switch/exchange
void Population::crossover(bool lc)
{
	crossover(0, size() - 1, lc);
}

// Crossover for Individuals between min and max indices, inclusive
void Population::crossover(int min, int max, bool lc)
{
	// array to hold the result of each crossover
	// array to hold the children
	vector<Individual> result;

	// make sure indices are OK
	int tempi = 0;
	if(min < 0)
		min = 0;
	else if(max > this->size() - 1)
		max = this->size() - 1;

	if(max < min)
	{
		tempi = max;
		max = min;
		min = tempi;
	}

	// calculate the number of parents involved in the crossover
	// then the number of children to be generated is also known
	int parents = max - min + 1;

	vector<Individual> children;

	// if the number of parents is odd, then one of them will cross twice
	// let the first one do an initial cross with the last one and then
	// remove the last one from the list
	bool odd = ((parents % 2) == 1);
	if(odd)
	{
		if(lc)
			result=((pool[min]) * (pool[max]));
		else
			result=((pool[min]) + (pool[max]));

		children.push_back(result[0]);
		children.push_back(result[1]);

		max--;
		result.clear();
	}

	// now there remain an even number of parents to be generated
	// cross them two at a time from index min to index max
	// note loop stops at i = max - 1, but since crossover includes
	// parent at index i + 1, the parent at max is included
	for(int i = min; i < max; i = i + 2)
	{
		if(lc)
			result=((pool[i]) * (pool[i + 1]));
		else
			result=((pool[i]) + (pool[i + 1]));

		children.push_back(result[0]);
		children.push_back(result[1]);

		result.clear();
	}

	// copy the children into the new pool
	for(int i = 0; i < children.size(); i++)
		pool.push_back(children.at(i));

	update();
}

// Offspring generation for all individuals
void Population::newOffspring(bool lc)
{
	newOffspring(0, size() - 1, lc);
}

// Offspring generation for individuals between min and max indices
void Population::newOffspring(int min, int max, bool lc)
{
	// array to hold the result of each crossover
	// array to hold the children
	vector<Individual> result;

	// make sure indices are OK
	int tempi = 0;
	if(min < 0)
		min = 0;
	else if(max > this->size() - 1)
		max = this->size() - 1;

	if(max < min)
	{
		tempi = max;
		max = min;
		min = tempi;
	}

	// calculate the number of parents involved in the crossover
	// then the number of children to be generated is also known
	int parents = max - min + 1;

	vector<Individual> children;

	// if the number of parents is odd, then one of them will cross twice
	// let the first one do an initial cross with the last one and then
	// remove the last one from the list
	bool odd = ((parents % 2) == 1);
	if(odd)
	{
		if(lc)
			result=(pool[min]) * (pool[max]);
		else
			result=(pool[min]) + (pool[max]);

		children.push_back(result[0]);
		children.push_back(result[1]);

		max--;
		result.clear();
	}

	// now there remain an even number of parents to be generated
	// cross them two at a time from index min to index max
	// note loop stops at i = max - 1, but since crossover includes
	// parent at index i + 1, the parent at max is included
	for(int i = min; i < max; i = i + 2)
	{
		if(lc)
			result=(pool[i]) * (pool[i + 1]);
		else
			result=(pool[i]) + (pool[i + 1]);

		children.push_back(result[0]);
		children.push_back(result[1]);

		result.clear();
	}

	// mutate and copy the children into the new pool
	for(int i = 0; i < children.size(); i++)
	{
		!children[i];
		pool.push_back(children[i]);
	}

	update();
}

// Selection function - calls roulette or tournament depending on the input
void Population::select(int samplesize, bool ts)
{
	setElite(fittest());

	if(ts)
		tournament(samplesize);
	else
		roulette();
}

// Set mutation rate for entire Population
void Population::setMR(double mr)
{
	setMR(mr, 0, size() - 1);
}

// Set mutation rate for specific Individuals
void Population::setMR(double mr, int min, int max)
{
	int tempi = 0;
	if(min < 0)
		min = 0;
	else if(max > this->size() - 1)
		max = this->size() - 1;

	if(min < max)
	{
		tempi = max;
		max = min;
		min = tempi;
	}

	for(int i = min; i <= max; i++)
		pool.at(i).setMR(mr);
}

// Set permutation for entire Population
void Population::setPR(double pr)
{
	setPR(pr, 0, size() - 1);
}

// Set permutation for specific Individuals
void Population::setPR(double pr, int min, int max)
{
	int tempi = 0;
	if(min < 0)
		min = 0;
	else if(max > this->size() - 1)
		max = this->size() - 1;

	if(min < max)
	{
		tempi = max;
		max = min;
		min = tempi;
	}

	for(int i = min; i <= max; i++)
		pool.at(i).setPR(pr);
}

// Set SD factor for entire Population
void Population::setFactor(double sdf)
{
	setFactor(sdf, 0, size() - 1);
}

// Set SD factor for specified Individuals
void Population::setFactor(double sdf, int min, int max)
{
	int tempi = 0;
	if(min < 0)
		min = 0;
	else if(max > this->size() - 1)
		max = this->size() - 1;

	if(min < max)
	{
		tempi = max;
		max = min;
		min = tempi;
	}

	for(int i = min; i <= max; i++)
		pool.at(i).setSDF(sdf);
}

// Find the fittest Individual in the specified group of Individuals
Individual& Population::fittest(vector<Individual> group)
{
	int best = 0;
	for(int i = 0; i < group.size(); i++)
	{
		if(pool[i].fitness() < pool[best].fitness())
			best = i;
	}

	return pool[best];
}

// Find the fittest Individual in the Population
Individual& Population::fittest()
{
	return fittest(pool);
}

// return mean fitness
double Population::fitness()
{
	return _fitness;
}

// return population size
int Population::size()
{
	return pool.size();
}

// return mean age
double Population::age()
{
	return _age;
}

// return standard deviation of population fitness
double Population::stddev()
{
	return _stddev;
}

// return relative standard deviation
double Population::rsd()
{
	return _rsd;
}

// return the number of elites
int Population::numElites()
{
	return elites.size();
}

// return the elite status of the desired Individual
bool Population::isElite(Individual& potential)
{
	return potential.isElite();
}

// check whether a solution exists
bool Population::solution()
{
	bool found = false;

	if(fittest().fitness() < tolerance)
		found = true;

	return found;
}

// return a nicely formatted string of information
string Population::info()
{
	stringstream output;
	output << "Population Statistics" << endl;
	output << "Size: " << size() << endl;
	output << "Mean Fitness: " << fitness() << endl;
	output << "Standard Deviation: " << stddev() << endl;
	output << "Mean Age: " << age() << endl;
	output << endl;
	output << "Fittest Individual" << endl;
	output << "Genotype: " << fittest().info() << endl;
	output << "Fitness: " << fittest().fitness() << endl;
	return output.str();
}





